@@ -1,4 +1,4 @@
-# Copyright 2000-2004 The Apache Software Foundation
+# Copyright 2000-2005 The Apache Software Foundation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -26,17 +26,19 @@ use ExtUtils::Embed ();
 
 use constant IS_MOD_PERL_BUILD => grep { -e "$_/lib/mod_perl.pm" } qw(. ..);
 
-use constant AIX    => $^O eq 'aix';
-use constant DARWIN => $^O eq 'darwin';
-use constant HPUX   => $^O eq 'hpux';
+use constant AIX     => $^O eq 'aix';
+use constant DARWIN  => $^O eq 'darwin';
+use constant IRIX    => $^O eq 'irix';
+use constant HPUX    => $^O eq 'hpux';
 use constant OPENBSD => $^O eq 'openbsd';
-use constant WIN32  => $^O eq 'MSWin32';
+use constant WIN32   => $^O eq 'MSWin32';
 
 use constant MSVC => WIN32() && ($Config{cc} eq 'cl');
 
 use constant REQUIRE_ITHREADS => grep { $^O eq $_ } qw(MSWin32);
 use constant PERL_HAS_ITHREADS =>
     $Config{useithreads} && ($Config{useithreads} eq 'define');
+use constant BUILD_APREXT => WIN32();
 
 use ModPerl::Code ();
 use ModPerl::BuildOptions ();
@@ -80,7 +82,7 @@ sub ap_prefix_invalid {
 
     my $include_dir = $self->apxs(-q => 'INCLUDEDIR');
 
-    unless (-e $include_dir) {
+    unless (-d $include_dir) {
         return "include/ directory not found in $prefix";
     }
 
@@ -151,7 +153,7 @@ sub apxs {
     my $query_key;
     if ($is_query) {
         $query_key = 'APXS_' . uc $_[1];
-        if ($self->{$query_key}) {
+        if (exists $self->{$query_key}) {
             return $self->{$query_key};
         }
     }
@@ -165,12 +167,12 @@ sub apxs {
 
     my $devnull = devnull();
     my $val = qx($apxs @_ 2>$devnull);
-    chomp $val if defined $val; # apxs post-2.0.40 adds a new line
+    chomp $val if defined $val;
 
     unless ($val) {
         # do we have an error or is it just an empty value?
         my $error = qx($apxs @_ 2>&1);
-        chomp $error if defined $error; # apxs post-2.0.40 adds a new line
+        chomp $error if defined $error;
         if ($error) {
             error "'$apxs @_' failed:";
             error $error;
@@ -184,23 +186,31 @@ sub apxs {
 }
 
 sub apxs_cflags {
-    my $cflags = __PACKAGE__->apxs('-q' => 'CFLAGS');
+    my $who = caller_package(shift);
+    my $cflags = $who->apxs('-q' => 'CFLAGS');
     $cflags =~ s/\"/\\\"/g;
     $cflags;
 }
 
 sub apxs_extra_cflags {
-    my $flags = __PACKAGE__->apxs('-q' => 'EXTRA_CFLAGS');
+    my $who = caller_package(shift);
+    my $flags = $who->apxs('-q' => 'EXTRA_CFLAGS');
     $flags =~ s/\"/\\\"/g;
     $flags;
 }
 
 sub apxs_extra_cppflags {
-    my $flags = __PACKAGE__->apxs('-q' => 'EXTRA_CPPFLAGS');
+    my $who = caller_package(shift);
+    my $flags = $who->apxs('-q' => 'EXTRA_CPPFLAGS');
     $flags =~ s/\"/\\\"/g;
     $flags;
 }
 
+sub caller_package {
+    my $arg = shift;
+    return ($arg and ref($arg) eq __PACKAGE__) ? $arg : __PACKAGE__;
+}
+
 my %threaded_mpms = map { $_ => 1}
         qw(worker winnt beos mpmt_os2 netware leader perchild threadpool);
 sub mpm_is_threaded {
@@ -238,12 +248,70 @@ sub mpm_name {
         $msg .= " Please specify MP_APXS=/full/path/to/apxs to solve " .
             "this problem." unless exists $self->{MP_APXS};
         error $msg;
-        exit 1;
+        die "\n";
     }
 
     return $self->{mpm_name} = $mpm_name;
 }
 
+sub should_build_apache {
+    my ($self) = @_;
+    return $self->{MP_USE_STATIC} ? 1 : 0;
+}
+
+sub configure_apache {
+    my ($self) = @_;
+
+    unless ($self->{MP_AP_CONFIGURE}) {
+        error "You specified MP_USE_STATIC but did not specify the " .
+              "arguments to httpd's ./configure with MP_AP_CONFIGURE";
+        exit 1;
+    }
+
+    unless ($self->{MP_AP_PREFIX}) {
+        error "You specified MP_USE_STATIC but did not speficy the " .
+              "location of httpd's source tree with MP_AP_PREFIX"; 
+        exit 1;
+    }
+
+    debug "Configuring httpd in $self->{MP_AP_PREFIX}";
+
+    my $httpd = File::Spec->catfile($self->{MP_AP_PREFIX}, 'httpd');
+    $self->{'httpd'} ||= $httpd;
+    push @Apache::TestMM::Argv, ('httpd' => $self->{'httpd'});
+
+    my $mplib = "$self->{MP_LIBNAME}$Config{lib_ext}";
+    my $mplibpath = catfile($self->{cwd}, qw(src modules perl), $mplib);
+
+    local $ENV{BUILTIN_LIBS} = $mplibpath;
+    local $ENV{AP_LIBS} = $self->ldopts;
+    local $ENV{MODLIST} = 'perl';
+
+    #XXX: -Wall and/or -Werror at httpd configure time breaks things
+    local $ENV{CFLAGS} = join ' ', grep { ! /\-Wall|\-Werror/ } 
+        split /\s+/, $ENV{CFLAGS} || '';
+
+    my $cd = qq(cd $self->{MP_AP_PREFIX});
+
+    #We need to clean the httpd tree before configuring it
+    if (-f File::Spec->catfile($self->{MP_AP_PREFIX}, 'Makefile')) {
+        my $cmd = qq(make clean);
+        debug "Running $cmd";
+        system("$cd && $cmd") == 0 or die "httpd: $cmd failed";
+    }
+
+    my $cmd = qq(./configure $self->{MP_AP_CONFIGURE});
+    debug "Running $cmd";
+    system("$cd && $cmd") == 0 or die "httpd: $cmd failed";
+
+    # Got to build in srclib/* early to have generated files present.
+    my $srclib = File::Spec->catfile($self->{MP_AP_PREFIX}, 'srclib');
+    $cd = qq(cd $srclib);
+    $cmd = qq(make);
+    debug "Building srclib in $srclib";
+    system("$cd && $cmd") == 0 or die "srclib: $cmd failed";
+}
+
 #--- Perl Config stuff ---
 
 my %gtop_config = ();
@@ -268,6 +336,7 @@ sub find_gtop {
 sub find_gtop_config {
     my %c = ();
 
+    my $ver_2_5_plus = 0;
     if (system('pkg-config --exists libgtop-2.0') == 0) {
         # 2.x
         chomp($c{ccopts} = qx|pkg-config --cflags libgtop-2.0|);
@@ -279,7 +348,9 @@ sub find_gtop_config {
 
         chomp($c{ver} = qx|pkg-config --modversion libgtop-2.0|);
         ($c{ver_maj}, $c{ver_min}) = split /\./, $c{ver};
-        if ($c{ver_maj} == 2 && $c{ver_min} >= 5) {
+        $ver_2_5_plus++ if $c{ver_maj} == 2 && $c{ver_min} >= 5;
+
+        if ($ver_2_5_plus) {
             # some headers were removed in libgtop 2.5.0 so we need to
             # be able to exclude them at compile time
             $c{ccopts} .= ' -DGTOP_2_5_PLUS';
@@ -295,7 +366,9 @@ sub find_gtop_config {
         $c{ldopts} =~ s|^/|-L/|;
     }
 
-    if ($c{ccopts}) {
+    # starting from 2.5.0 'pkg-config --cflags libgtop-2.0' already
+    # gives us all the cflags that are needed
+    if ($c{ccopts} && !$ver_2_5_plus) {
         chomp(my $ginc = `glib-config --cflags`);
         $c{ccopts} .= " $ginc";
     }
@@ -370,6 +443,13 @@ sub ldopts {
 
     $config->{ldflags} = $ldflags; #reset
 
+    # on Irix mod_perl.so needs to see the libperl.so symbols, which
+    # requires the -exports option immediately before -lperl.
+    if (IRIX) {
+        ($ldopts =~ s/-lperl\b/-exports -lperl/)
+            or warn "Failed to fix Irix symbol exporting\n";
+    }
+
     $ldopts;
 }
 
@@ -693,11 +773,13 @@ sub new {
     my $class = shift;
 
     my $self = bless {
-        cwd => Cwd::fastcwd(),
+        cwd        => Cwd::fastcwd(),
         MP_LIBNAME => 'mod_perl',
         @_,
     }, $class;
 
+    $self->{MP_APR_LIB} = 'aprext';
+
     ModPerl::BuildOptions->init($self) if delete $self->{init};
 
     $self;
@@ -707,8 +789,8 @@ sub DESTROY {}
 
 my %default_files = (
     'build_config' => 'lib/Apache/BuildConfig.pm',
-    'ldopts' => 'src/modules/perl/ldopts',
-    'makefile' => 'src/modules/perl/Makefile.modperl',
+    'ldopts'       => 'src/modules/perl/ldopts',
+    'makefile'     => 'src/modules/perl/Makefile',
 );
 
 sub clean_files {
@@ -730,7 +812,8 @@ sub file_path {
 
 sub freeze {
     require Data::Dumper;
-    local $Data::Dumper::Terse = 1;
+    local $Data::Dumper::Terse    = 1;
+    local $Data::Dumper::Sortkeys = 1;
     my $data = Data::Dumper::Dumper(shift);
     chomp $data;
     $data;
@@ -760,7 +843,10 @@ sub save {
     $file ||= $self->default_file('build_config');
     $file = $self->file_path($file);
 
-    (my $obj = $self->freeze) =~ s/^/    /;
+    my $obj = $self->freeze;
+    $obj =~ s/^\s{9}//mg;
+    $obj =~ s/^/    /;
+
     open my $fh, '>', $file or die "open $file: $!";
 
     #work around autosplit braindeadness
@@ -876,6 +962,43 @@ sub ap_includedir  {
     $self->{ap_includedir} = $d;
 }
 
+# This is necessary for static builds that needs to make a
+# difference between where the apache headers are (to build
+# against) and where they will be installed (to install our
+# own headers alongside)
+# 
+# ap_exp_includedir is where apache is going to install its
+# headers to
+sub ap_exp_includedir {
+    my ($self) = @_;
+    
+    return $self->{ap_exp_includedir} if $self->{ap_exp_includedir};
+    
+    my $build_vars = File::Spec->catfile($self->{MP_AP_PREFIX}, 
+                                         qw(build config_vars.mk));
+    open my $vars, "<$build_vars" or die "Couldn't open $build_vars $!";
+    my $ap_exp_includedir;
+    while (<$vars>) {
+        if (/exp_includedir\s*=\s*(.*)/) {
+            $ap_exp_includedir = $1;
+            last;
+        }
+    }
+    
+    $self->{ap_exp_includedir} = $ap_exp_includedir;
+}
+
+sub install_headers_dir {
+    my ($self) = @_;
+    if ($self->should_build_apache) {
+        return $self->ap_exp_includedir();
+    }
+    else {
+        return $self->ap_includedir();
+    }
+}
+
+
 # where apr-config and apu-config reside
 sub apr_bindir {
     my ($self) = @_;
@@ -884,34 +1007,76 @@ sub apr_bindir {
     $self->{apr_bindir};
 }
 
-# XXX: we assume that apr-config and apu-config reside in the same
-# directory
-sub apr_config_path {
-    my ($self) = @_;
+sub apr_generation {
+    my($self) = @_;
+    return $self->httpd_version_as_int =~ m/21\d+/ ? 1 : 0;
+}
 
-    return $self->{apr_config_path}
-        if $self->{apr_config_path} and -x $self->{apr_config_path};
+# returns an array of apr/apu linking flags (--link-ld --libs) if found
+# an empty array otherwise
+my @apru_link_flags = ();
+sub apru_link_flags {
+    my($self) = @_;
 
-    if (exists $self->{MP_APR_CONFIG} and -x $self->{MP_APR_CONFIG}) {
-        $self->{apr_config_path} = $self->{MP_APR_CONFIG};
+    return @apru_link_flags if @apru_link_flags;
+
+    for ($self->apr_config_path, $self->apu_config_path) {
+        if (my $link = $_ && -x $_ && qx{$_ --link-ld --libs}) {
+            chomp $link;
+            if ($self->httpd_is_source_tree) {
+                my @libs;
+                while ($link =~ m/-L(\S+)/g) {
+                    my $dir = File::Spec->catfile($1, '.libs');
+                    push @libs, $dir if -d $dir;
+                }
+                push @apru_link_flags, join ' ', map { "-L$_" } @libs;
+            }
+            push @apru_link_flags, $link;
+        }
     }
 
-    if (!$self->{apr_config_path}) {
+    return @apru_link_flags;
+}
+
+sub apr_config_path {
+    shift->apru_config_path("apr");
+}
+
+sub apu_config_path {
+    shift->apru_config_path("apu");
+}
+
+sub apru_config_path {
+    my ($self, $what) = @_;
+
+    my $key = "${what}_config_path"; # apr_config_path
+    my $mp_key = "MP_" . uc($what) . "_CONFIG"; # MP_APR_CONFIG
+    my $bindir = uc($what) . "_BINDIR"; # APR_BINDIR
+
+    return $self->{$key} if $self->{$key} and -x $self->{$key};
+
+    if (exists $self->{$mp_key} and -x $self->{$mp_key}) {
+        $self->{$key} = $self->{$mp_key};
+    }
+
+    my $config = $self->apr_generation ? "$what-1-config" : "$what-config";
+
+    if (!$self->{$key}) {
         my @tries = ();
         if ($self->httpd_is_source_tree) {
-            push @tries, grep { -d $_ }
-                map catdir($_, "srclib", "apr"),
-                grep defined $_, $self->dir;
+            for my $base (grep defined $_, $self->dir) {
+                push @tries, grep -d $_,
+                    map catdir($base, "srclib", $_), qw(apr apr-util);
+            }
         }
         else {
-            # APR_BINDIR was added only at httpd-2.0.46
             push @tries, grep length,
-                map $self->apxs(-q => $_), qw(APR_BINDIR BINDIR);
+                map $self->apxs(-q => $_), $bindir, "BINDIR";
             push @tries, catdir $self->{MP_AP_PREFIX}, "bin"
                 if exists $self->{MP_AP_PREFIX} and -d $self->{MP_AP_PREFIX};
         }
 
-        @tries = map { catfile $_, "apr-config" } @tries;
+        @tries = map { catfile $_, $config } @tries;
         if (WIN32) {
             my $ext = '.bat';
             for (@tries) {
@@ -921,22 +1086,22 @@ sub apr_config_path {
 
         for my $try (@tries) {
             next unless -x $try;
-            $self->{apr_config_path} = $try;
+            $self->{$key} = $try;
         }
     }
 
-    $self->{apr_config_path} ||= Apache::TestConfig::which('apr-config');
+    $self->{$key} ||= Apache::TestConfig::which($config);
 
     # apr_bindir makes sense only if httpd/apr is installed, if we are
     # building against the source tree we can't link against
     # apr/aprutil libs
     unless ($self->httpd_is_source_tree) {
-        $self->{apr_bindir} = $self->{apr_config_path}
-            ? dirname $self->{apr_config_path}
+        $self->{apr_bindir} = $self->{$key}
+            ? dirname $self->{$key}
             : '';
         }
 
-    $self->{apr_config_path};
+    $self->{$key};
 }
 
 sub apr_includedir {
@@ -965,6 +1130,7 @@ sub apr_includedir {
         push @tries, $path if -d $path;
     }
 
+
     for (@tries) {
         next unless $_ && -e catfile $_, "apr.h";
         $self->{apr_includedir} = $_;
@@ -1209,6 +1375,8 @@ EOF
 
 #when we use a bit of MakeMaker, make it use our values for these vars
 my %perl_config_pm_alias = (
+    ABSPERL      => 'perlpath',
+    ABSPERLRUN   => 'perlpath',
     PERL         => 'perlpath',
     PERLRUN      => 'perlpath',
     PERL_LIB     => 'privlibexp',
@@ -1217,9 +1385,12 @@ my %perl_config_pm_alias = (
 
 my $mm_replace = join '|', keys %perl_config_pm_alias;
 
-my @perl_config_pm =
-  (qw(cc cpprun rm ranlib lib_ext obj_ext cccdlflags lddlflags optimize),
-   values %perl_config_pm_alias);
+# get rid of dups
+my %perl_config_pm_alias_values = reverse %perl_config_pm_alias;
+my @perl_config_pm_alias_values = keys %perl_config_pm_alias_values;
+
+my @perl_config_pm = (@perl_config_pm_alias_values, qw(cc cpprun
+    rm ranlib lib_ext obj_ext cccdlflags lddlflags optimize));
 
 sub mm_replace {
     my $val = shift;
@@ -1344,6 +1515,36 @@ sub modperl_libs {
     $libs->($self);
 }
 
+# returns the directory and name of the aprext lib built under blib/ 
+sub mp_apr_blib {
+    my $self = shift;
+    return unless (my $mp_apr_lib = $self->{MP_APR_LIB});
+    my $lib_mp_apr_lib = 'lib' . $mp_apr_lib;
+    my @dirs = $self->{MP_INST_APACHE2} ?
+        qw(blib arch Apache2 auto) : qw(blib arch auto);
+    my $apr_blib = catdir $self->{cwd}, @dirs, $lib_mp_apr_lib;
+    my $full_libname = $lib_mp_apr_lib . $Config{lib_ext};
+    return ($apr_blib, $full_libname);
+}
+
+sub mp_apr_lib_MSWin32 {
+    my $self = shift;
+    # The MP_APR_LIB will be installed into MP_AP_PREFIX/lib
+    # for use by 3rd party xs modules
+    my ($dir, $lib) = $self->mp_apr_blib();
+    $lib =~ s[^lib(\w+)$Config{lib_ext}$][$1];
+    $dir = Win32::GetShortPathName($dir);
+    return qq{ -L$dir -l$lib };
+}
+
+# linking used for the aprext lib used to build APR/APR::*
+sub mp_apr_lib {
+    my $self = shift;
+    my $libs = \&{"mp_apr_lib_$^O"};
+    return "" unless defined &$libs;
+    $libs->($self);
+}
+
 sub modperl_symbols_MSWin32 {
     my $self = shift;
     return "" unless $self->{MP_DEBUG};
@@ -1364,10 +1565,17 @@ sub write_src_makefile {
 
     my $install = <<'EOI';
 install:
+EOI
+    if (!$self->should_build_apache) {
+        $install .= <<'EOI';
 # install mod_perl.so
 	@$(MKPATH) $(MODPERL_AP_LIBEXECDIR)
 	$(MODPERL_TEST_F) $(MODPERL_LIB_DSO) && \
 	$(MODPERL_CP) $(MODPERL_LIB_DSO) $(MODPERL_AP_LIBEXECDIR)
+EOI
+    }
+    
+    $install .= <<'EOI';
 # install mod_perl .h files
 	@$(MKPATH) $(MODPERL_AP_INCLUDEDIR)
 	$(MODPERL_CP) $(MODPERL_H_FILES) $(MODPERL_AP_INCLUDEDIR)
@@ -1455,11 +1663,12 @@ EOI
     }
 
     print $fh $self->canon_make_attr('lib', "@libs");
+    
+    print $fh $self->canon_make_attr('AP_INCLUDEDIR', 
+                                     $self->install_headers_dir());
 
-    for my $q (qw(LIBEXECDIR INCLUDEDIR)) {
-        print $fh $self->canon_make_attr("AP_$q",
-                                         $self->apxs(-q => $q));
-    }
+    print $fh $self->canon_make_attr('AP_LIBEXECDIR',
+                                     $self->apxs(-q => 'LIBEXECDIR'));
 
     my $xs_targ = $self->make_xs($fh);
 
@@ -1588,12 +1797,14 @@ sub includes {
         my $ap_inc = $self->apxs('-q' => 'INCLUDEDIR');
         if ($ap_inc && -d $ap_inc) {
             push @inc, $ap_inc;
-        } else {
-            # this is fatal
-            die "Can't find the mod_perl include dir";
+            return \@inc;
         }
 
-        return \@inc;
+        # this is fatal
+        my $reason = $ap_inc
+            ? "path $ap_inc doesn't exist"
+            : "apxs -q INCLUDEDIR didn't return a value";
+        die "Can't find the mod_perl include dir (reason: $reason)";
     }
 
     my $src = $self->dir;
@@ -1605,6 +1816,14 @@ sub includes {
     unless ($self->httpd_is_source_tree) {
         push @inc, $self->apr_includedir;
 
+        my $apuc = $self->apu_config_path;
+        if ($apuc && -x $apuc) {
+            chomp(my $apuincs = qx($apuc --includes));
+            # win32: /Ipath, elsewhere -Ipath
+            $apuincs =~ s{^\s*(-|/)I}{};
+            push @inc, $apuincs;
+        }
+
         my $ainc = $self->apxs('-q' => 'INCLUDEDIR');
         if (-d $ainc) {
             push @inc, $ainc;
