@@ -1,4 +1,4 @@
-# Copyright 2003-2004 The Apache Software Foundation
+# Copyright 2003-2005 The Apache Software Foundation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,12 +17,12 @@ package Apache::PerlSections;
 use strict;
 use warnings FATAL => 'all';
 
-our $VERSION = '0.01';
+our $VERSION = '2.00';
 
 use Apache::CmdParms ();
 use Apache::Directive ();
 use APR::Table ();
-use Apache::Server ();
+use Apache::ServerRec ();
 use Apache::ServerUtil ();
 use Apache::Const -compile => qw(OK);
 
@@ -34,12 +34,12 @@ sub new {
     return bless { @args }, ref($package) || $package;
 }
 
-sub server     { return shift->{'parms'}->server() }
+sub parms      { return shift->{'parms'} }
 sub directives { return shift->{'directives'} ||= [] }
 sub package    { return shift->{'args'}->{'package'} }
 
 my @saved;
-sub save       { return $Apache::Server::SaveConfig }
+sub save       { return $Apache::PerlSections::Save }
 sub saved      { return @saved }
 
 sub handler : method {
@@ -79,9 +79,9 @@ sub symdump {
 
     unless ($self->{symbols}) {
         no strict;
-        
+
         $self->{symbols} = [];
-        
+
         #XXX: Here would be a good place to warn about NOT using 
         #     Apache::ReadConfig:: directly in <Perl> sections
         foreach my $pack ($self->package, $self->SPECIAL_PACKAGE) {
@@ -172,7 +172,12 @@ sub dump_entry {
         $self->add_config("$name $$entry\n");
     }
     elsif ($type eq 'ARRAY') {
-        $self->add_config("$name @$entry\n");
+        if (grep {ref} @$entry) {
+            $self->dump_entry($name, $_) for @$entry;
+        }
+        else {
+            $self->add_config("$name @$entry\n");
+        }
     }
     elsif ($type eq 'HASH') {
         $self->dump_hash($name, $entry);
@@ -195,7 +200,7 @@ sub add_config {
 
 sub post_config {
     my($self) = @_;
-    my $errmsg = $self->server->add_config($self->directives);
+    my $errmsg = $self->parms->add_config($self->directives);
     die $errmsg if $errmsg;
 }
 
