@@ -1,4 +1,4 @@
-# Copyright 2003-2004 The Apache Software Foundation
+# Copyright 2003-2005 The Apache Software Foundation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -67,7 +67,7 @@ delete $status{'hooks'} if $mod_perl::VERSION >= 1.9901;
 delete $status{'sig'} if IS_WIN32;
 
 # XXX: needs porting
-if ($Apache::Server::SaveConfig) {
+if ($Apache::PerlSections::Save) {
     $status{"section_config"} = "Perl Section Configuration";
 }
 
@@ -99,7 +99,7 @@ sub has {
     # if !$opt we skip the testing for the option
     return 0 if $opt && !status_config($r, $opt);
     return 0 unless eval { require $file };
-    return 0 unless $module->VERSION >= $version;
+    return 0 unless $module->VERSION && $module->VERSION >= $version;
 
     return 1;
 }
@@ -161,7 +161,7 @@ sub handler {
 sub header {
     my $r = shift;
     my $start = scalar localtime $^T;
-    my $srv = Apache::get_server_version();
+    my $srv = Apache::ServerUtil::get_server_version();
     $r->content_type("text/html");
     my $v = $^V ? sprintf "v%vd", $^V : $];
     $r->print(<<"EOF");
@@ -350,7 +350,7 @@ sub status_env {
     if ($r->handler eq 'modperl') {
         # the handler can be executed under the "modperl" handler
         push @retval,
-            qq{<b>Under the "modperl" handler, the environment is:</b>};
+            qq{<b>Under the "modperl" handler, the environment is</b>:};
         # XXX: I guess we could call $r->subprocess_env; and show how
         # would it look like under the 'perl-script' environment, but
         # under the 'modperl' handler %ENV doesn't get reset,
@@ -362,7 +362,9 @@ sub status_env {
             qq{<b>Under the "perl-script" handler, the environment is</b>:};
     }
     push @retval, "\n</p>\n";
-    push @retval, "<pre>", (map "$_ = $ENV{$_}\n", sort keys %ENV), "</pre>";
+    push @retval, "<pre>",
+        (map "$_ = " . escape_html($ENV{$_}||'') . "\n",
+            sort keys %ENV), "</pre>";
 
     \@retval;
 }
@@ -408,6 +410,7 @@ sub status_data_dump {
     no strict 'refs';
     my @retval = "<p>\nData Dump of $name $type\n</p>\n<pre>\n";
     my $str = Data::Dumper->Dump([*$name{$type}], ['*'.$name]);
+    $str = escape_html($str);
     $str =~ s/= \\/= /; #whack backwack
     push @retval, $str, "\n";
     push @retval, peek_link($r, $q, $name, $type);
@@ -463,7 +466,7 @@ sub b_lexinfo_link {
 
     return unless has($r, "lexinfo");
 
-    my $script = $q->location;
+    my $script = $r->location;
     return qq(\n<a href="$script/$name?noh_b_lexinfo">Lexical Info</a>\n);
 }
 
@@ -704,8 +707,8 @@ sub noh_b_graph {
 
     untie *STDOUT;
 
-    my $dir = 
-        $r->server_root_relative($r->dir_config("GraphDir") || "logs/b_graphs");
+    my $dir = File::Spec->catfile(Apache::ServerUtil::server_root(),
+        ($r->dir_config("GraphDir") || "logs/b_graphs"));
 
     mkdir $dir, 0755 unless -d $dir;
 
@@ -721,10 +724,8 @@ sub noh_b_graph {
     }
 
     if (-s $file) {
-        local *FH;
-        open FH, $file or die "Can't open $file: $!";
         $r->content_type("image/gif");
-        $r->send_fd(\*FH);
+        $r->sendfile($file);
     }
     else {
         $r->content_type("text/plain");
@@ -827,6 +828,16 @@ sub as_HTML {
     return join "\n", @m, "<hr>", b_package_size_link($r, $q, $package);
 }
 
+sub escape_html {
+    my $str = shift;
+
+    $str =~ s/&/&amp;/g;
+    $str =~ s/</&lt;/g;
+    $str =~ s/>/&gt;/g;
+
+    return $str;
+}
+
 sub myconfig {
     require Config;
     # Config::myconfig(); fails under threads with (5.8.0 < perl < 5.8.3)
