@@ -1,3 +1,18 @@
+/* Copyright 2000-2005 The Apache Software Foundation
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
+
 #include "mod_perl.h"
 
 /* make sure that mod_perl won't try to start itself, while it's
@@ -9,6 +24,33 @@ static int MP_init_status = 0;
 #define MP_IS_STARTING    (MP_init_status == 1 ? 1 : 0)
 #define MP_IS_RUNNING     (MP_init_status == 2 ? 1 : 0)
 
+/* false while there is only the parent process and may be child
+ * processes, but no threads around, useful for allowing things that
+ * don't require locking and won't affect other threads. It should
+ * become true just before the child_init phase  */
+static int MP_threads_started = 0;
+
+int modperl_threads_started(void)
+{
+    return MP_threads_started;
+}
+
+static int MP_threaded_mpm = 0;
+
+int modperl_threaded_mpm(void)
+{
+    return MP_threaded_mpm;
+}
+
+/* sometimes non-threaded mpm also needs to know whether it's still
+ * starting up or after post_config) */
+static int MP_post_post_config_phase = 0;
+
+int modperl_post_post_config_phase(void)
+{
+    return MP_post_post_config_phase;
+}
+
 #ifndef USE_ITHREADS
 static apr_status_t modperl_shutdown(void *data)
 {
@@ -70,13 +112,12 @@ static void modperl_boot(pTHX_ void *data)
     /* outside mod_perl this is done by ModPerl::Const.xs */
     newXS("ModPerl::Const::compile", XS_modperl_const_compile, __FILE__);
 
-#ifdef MP_PERL_5_6_x
     /* make sure DynaLoader is loaded before XSLoader
-     * to workaround bug in 5.6.1 that can trigger a segv
+     * - to workaround bug in 5.6.1 that can trigger a segv
      * when using modperl as a dso
+     * - also needed when <Perl> sections are loaded from +Parent vhost
      */
     modperl_require_module(aTHX_ "DynaLoader", FALSE);
-#endif
 
     IoFLUSH_on(PL_stderrgv); /* unbuffer STDERR */
 }
@@ -98,14 +139,26 @@ static void modperl_xs_init(pTHX)
  * the parent process will run the cleanups since server_pool is a subpool
  * of pconf.  we manually clear the server_pool to run cleanups in the
  * child processes
+ *
+ * the "server_user_pool" is a subpool of the "server_pool", this is
+ * the pool which is exposed to users, so that they can register
+ * cleanup callbacks. This is needed so that the perl cleanups won't
+ * be run before user cleanups are executed.
+ *
  */
 static apr_pool_t *server_pool = NULL;
+static apr_pool_t *server_user_pool = NULL;
 
 apr_pool_t *modperl_server_pool(void)
 {
     return server_pool;
 }
 
+apr_pool_t *modperl_server_user_pool(void)
+{
+    return server_user_pool;
+}
+
 static void set_taint_var(PerlInterpreter *perl)
 {
     dTHXa(perl);
@@ -127,7 +180,7 @@ static void set_taint_var(PerlInterpreter *perl)
         SvREADONLY_on(GvSV(gv));
     }
 #endif /* MP_COMPAT_1X */
-    
+
 }
 
 PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
@@ -149,10 +202,10 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
     if (MP_init_status != 2) {
         server_rec *base_server = modperl_global_get_server_rec();
         PerlInterpreter *base_perl;
-        
+
         MP_init_status = 2; /* calls itself, so set the flag early */
         base_perl = modperl_startup(base_server, p);
-        
+
         if (base_server == s ) {
             return base_perl;
         }
@@ -217,7 +270,7 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
     /* workaround perl5.8.0/glibc bug */
     PL_reentrant_buffer->_crypt_struct.current_saltbits = 0;
 #endif
-    
+
     perl_run(perl);
 
 #ifdef USE_ITHREADS
@@ -231,7 +284,7 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
     PL_endav = endav;
 
     set_taint_var(perl);
-    
+
     MP_TRACE_i(MP_FUNC, "constructed interpreter=0x%lx\n",
                (unsigned long)perl);
 
@@ -242,12 +295,30 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
 #endif
 
 #ifdef MP_COMPAT_1X
-    av_push(GvAV(PL_incgv),
-            newSVpv(ap_server_root_relative(p, ""), 0));
-    av_push(GvAV(PL_incgv),
-            newSVpv(ap_server_root_relative(p, "lib/perl"), 0));
+    {
+        char *path, *path1;
+        apr_finfo_t finfo;
+        /* 1) push @INC, $ServerRoot */
+        av_push(GvAV(PL_incgv), newSVpv(ap_server_root, 0));
+
+        /* 2) push @INC, $ServerRoot/lib/perl only if it exists */
+        apr_filepath_merge(&path, ap_server_root, "lib",
+                           APR_FILEPATH_NATIVE, p);
+        apr_filepath_merge(&path1, path, "perl",
+                           APR_FILEPATH_NATIVE, p);
+        if (APR_SUCCESS == apr_stat(&finfo, path1, APR_FINFO_TYPE, p)) {
+            if (finfo.filetype == APR_DIR) {
+                av_push(GvAV(PL_incgv), newSVpv(path1, 0));
+            }
+        }
+    }
 #endif /* MP_COMPAT_1X */
-    
+
+    /* things to be done only in the main server */
+    if (!s->is_virtual) {
+        modperl_handler_anon_init(aTHX_ p);
+    }
+
     if (!modperl_config_apply_PerlRequire(s, scfg, perl, p)) {
         exit(1);
     }
@@ -285,7 +356,7 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
 
     MP_TRACE_i(MP_FUNC, "Init vhost %s: s=0x%lx, base_s=0x%lx\n",
                vhost, s, base_server);
-    
+
     if (base_server == s) {
         MP_TRACE_i(MP_FUNC, "base server is not vhost, skipping %s\n",
                    vhost);
@@ -307,7 +378,9 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
         return OK;
     }
 
-    if (!MpSrvENABLE(scfg)) {
+    /* the base server could have mod_perl callbacks disabled, but it
+     * still needs perl to drive the vhosts */
+    if (!MpSrvENABLE(scfg) && s->is_virtual) {
         MP_TRACE_i(MP_FUNC, "mod_perl disabled for server %s\n", vhost);
         scfg->mip = NULL;
         return OK;
@@ -333,7 +406,7 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
             modperl_interp_init(s, p, perl);
         }   
 #endif
-        
+
         if (!modperl_config_apply_PerlRequire(s, scfg, perl, p)) {
             return HTTP_INTERNAL_SERVER_ERROR;
         }
@@ -370,18 +443,13 @@ void modperl_init(server_rec *base_server, apr_pool_t *p)
                                    base_server->server_hostname));
 
 #ifndef USE_ITHREADS
-    if (base_scfg->threaded_mpm) {
+    if (modperl_threaded_mpm()) {
         ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
                      "cannot use threaded MPM without ithreads enabled Perl");
         exit(1);
     }
 #endif
 
-    if (!MpSrvENABLE(base_scfg)) {
-        /* how silly */
-        return;
-    }
-
     base_perl = modperl_startup(base_server, p);
 
     for (s=base_server->next; s; s=s->next) {
@@ -398,6 +466,17 @@ void modperl_init(server_rec *base_server, apr_pool_t *p)
 
 }
 
+static int modperl_post_config_require(server_rec *s, apr_pool_t *p)
+{
+    for (; s; s=s->next) {
+        MP_dSCFG(s);
+        if (!modperl_config_apply_PerlPostConfigRequire(s, scfg, p)) {
+            return FALSE;   
+        }
+    }
+    return TRUE;
+}
+
 #ifdef USE_ITHREADS
 static void modperl_init_clones(server_rec *s, apr_pool_t *p)
 {
@@ -406,7 +485,7 @@ static void modperl_init_clones(server_rec *s, apr_pool_t *p)
     char *base_name = modperl_server_desc(s, p);
 #endif /* MP_TRACE */
 
-    if (!base_scfg->threaded_mpm) {
+    if (!modperl_threaded_mpm()) {
         MP_TRACE_i(MP_FUNC, "no clones created for non-threaded mpm\n");
         return;
     }
@@ -447,13 +526,11 @@ static void modperl_init_clones(server_rec *s, apr_pool_t *p)
 
 void modperl_init_globals(server_rec *s, apr_pool_t *pconf)
 {
-    int threaded_mpm;
-    ap_mpm_query(AP_MPMQ_IS_THREADED, &threaded_mpm);
+    ap_mpm_query(AP_MPMQ_IS_THREADED, &MP_threaded_mpm);
 
     MP_TRACE_g(MP_FUNC, "mod_perl globals are configured\n");
-    
+
     modperl_global_init_pconf(pconf, pconf);
-    modperl_global_init_threaded_mpm(pconf, threaded_mpm);
     modperl_global_init_server_rec(pconf, s);
 
     modperl_tls_create_request_rec(pconf);
@@ -465,6 +542,8 @@ void modperl_init_globals(server_rec *s, apr_pool_t *pconf)
  */
 static apr_status_t modperl_sys_init(void)
 {
+    MP_TRACE_i(MP_FUNC, "mod_perl sys init\n");
+
 #if 0 /*XXX*/
     PERL_SYS_INIT(0, NULL);
 
@@ -491,6 +570,12 @@ static apr_status_t modperl_sys_init(void)
 static apr_status_t modperl_sys_term(void *data)
 {
     MP_init_status = 0;
+    MP_threads_started = 0;
+    MP_post_post_config_phase = 0;
+
+    MP_PERL_FREE_THREAD_KEY_WORKAROUND;
+
+    MP_TRACE_i(MP_FUNC, "mod_perl sys term\n");
 
     modperl_env_unload();
 
@@ -509,12 +594,20 @@ int modperl_hook_init(apr_pool_t *pconf, apr_pool_t *plog,
         return OK;
     }
 
+    MP_TRACE_i(MP_FUNC, "mod_perl hook init\n");
+
     MP_init_status = 1; /* now starting */
 
+    modperl_restart_count_inc(s);
+
     apr_pool_create(&server_pool, pconf);
+    apr_pool_tag(server_pool, "mod_perl server pool");
+
+    apr_pool_create(&server_user_pool, pconf);
+    apr_pool_tag(server_user_pool, "mod_perl server user pool");
 
     modperl_sys_init();
-    apr_pool_cleanup_register(pconf, NULL,
+    apr_pool_cleanup_register(server_pool, NULL,
                               modperl_sys_term, apr_pool_cleanup_null);
 
     modperl_init(s, pconf);
@@ -557,23 +650,45 @@ static int modperl_hook_pre_connection(conn_rec *c, void *csd)
     return OK;
 }
 
-static int modperl_hook_post_config(apr_pool_t *pconf, apr_pool_t *plog,
-                                    apr_pool_t *ptemp, server_rec *s)
+static int modperl_hook_post_config_last(apr_pool_t *pconf, apr_pool_t *plog,
+                                         apr_pool_t *ptemp, server_rec *s)
 {
+    /* in the threaded environment, no server_rec/process_rec
+     * modifications should be done beyond this point */
 #ifdef USE_ITHREADS
     MP_dSCFG(s);
     dTHXa(scfg->mip->parent->perl);
 #endif
 
+    if (!modperl_post_config_require(s, pconf)) {
+        exit(1);
+    }
+
+    if (modperl_threaded_mpm()) {
+        MP_threads_started = 1;
+    }
+
+    MP_post_post_config_phase = 1;
+
 #ifdef MP_TRACE
     /* httpd core open_logs handler re-opens s->error_log, which might
      * change, even though it still points to the same physical file
      * (.e.g on win32 the filehandle will be different. Therefore
      * reset the tracing logfile setting here, since this is the
-     * earliest place, happening after the open_logs phase */
-    modperl_trace_logfile_set(s->error_log);
+     * earliest place, happening after the open_logs phase.
+     *
+     * Moreover, we need to dup the filehandle so that when the server
+     * shuts down, we will be able to log to error_log after Apache
+     * has closed it (which happens too early for our likening).
+     */
+    {
+        apr_file_t *dup;
+        MP_RUN_CROAK(apr_file_dup(&dup, s->error_log, pconf),
+                     "mod_perl core post_config");
+        modperl_trace_logfile_set(dup);
+    }
 #endif
-    
+
     ap_add_version_component(pconf, MP_VERSION_STRING);
     ap_add_version_component(pconf,
                              Perl_form(aTHX_ "Perl/v%vd", PL_patchlevel));
@@ -588,7 +703,7 @@ static int modperl_hook_post_config(apr_pool_t *pconf, apr_pool_t *plog,
     ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
                  "mod_perl: using Perl HASH_SEED: %"UVuf, MP_init_hash_seed);
 #endif
-    
+
     return OK;
 }
 
@@ -604,7 +719,7 @@ static int modperl_hook_create_request(request_rec *r)
      * wasn't turned off and MpDirPARSE_HEADERS is on
      */
     MpReqPARSE_HEADERS_On(rcfg);
-    
+
     return OK;
 }
 
@@ -631,13 +746,26 @@ int modperl_perl_destruct_level(void)
     return modperl_destruct_level;
 }
 
+#ifdef USE_ITHREADS
+
+static apr_status_t
+modperl_perl_call_endav_mip(pTHX_ modperl_interp_pool_t *mip,
+                            void *data)
+{
+    modperl_perl_call_endav(aTHX);
+    return APR_SUCCESS;
+}
+
+#endif /* USE_ITHREADS */
+
 static apr_status_t modperl_child_exit(void *data)
 {
     char *level = NULL;
     server_rec *s = (server_rec *)data;
-    
-    modperl_callback_process(MP_CHILD_EXIT_HANDLER, server_pool, s, MP_HOOK_VOID);
-    
+
+    modperl_callback_process(MP_CHILD_EXIT_HANDLER, server_pool, s,
+                             MP_HOOK_VOID);
+
     if ((level = getenv("PERL_DESTRUCT_LEVEL"))) {
         modperl_destruct_level = atoi(level);
     }
@@ -649,6 +777,17 @@ static apr_status_t modperl_child_exit(void *data)
     if (modperl_destruct_level) {
         apr_pool_clear(server_pool);
     }
+    else {
+        /* run the END blocks of this child process if
+         * modperl_perl_destruct is not called for this process */
+#ifdef USE_ITHREADS
+        modperl_interp_mip_walk_servers(NULL, s,
+                                        modperl_perl_call_endav_mip,
+                                        (void*)NULL);
+#else
+        modperl_perl_call_endav(aTHX);
+#endif
+    }
 
     server_pool = NULL;
 
@@ -667,6 +806,11 @@ static void modperl_hook_child_init(apr_pool_t *p, server_rec *s)
 
 void modperl_register_hooks(apr_pool_t *p)
 {
+
+#ifdef USE_ITHREADS
+    APR_REGISTER_OPTIONAL_FN(modperl_interp_unselect);
+#endif
+
     /* for <IfDefine MODPERL2> and Apache->define("MODPERL2") */
     *(char **)apr_array_push(ap_server_config_defines) =
         apr_pstrdup(p, "MODPERL2");
@@ -677,8 +821,8 @@ void modperl_register_hooks(apr_pool_t *p)
     ap_hook_open_logs(modperl_hook_init,
                       NULL, NULL, APR_HOOK_FIRST);
 
-    ap_hook_post_config(modperl_hook_post_config,
-                        NULL, NULL, APR_HOOK_FIRST);
+    ap_hook_post_config(modperl_hook_post_config_last,
+                        NULL, NULL, APR_HOOK_REALLY_LAST);
 
     ap_hook_handler(modperl_response_handler,
                     NULL, NULL, APR_HOOK_MIDDLE);
@@ -714,11 +858,15 @@ void modperl_register_hooks(apr_pool_t *p)
     ap_hook_create_request(modperl_hook_create_request,
                            NULL, NULL, APR_HOOK_MIDDLE);
 
+    /* both of these hooks need to run really, really first.
+     * otherwise, the global request_rec will be set up _after_ some
+     * Perl handlers run.
+     */
     ap_hook_post_read_request(modperl_hook_post_read_request,
-                              NULL, NULL, APR_HOOK_FIRST);
+                              NULL, NULL, MODPERL_HOOK_REALLY_REALLY_FIRST);
 
     ap_hook_header_parser(modperl_hook_header_parser,
-                          NULL, NULL, APR_HOOK_FIRST);
+                          NULL, NULL, MODPERL_HOOK_REALLY_REALLY_FIRST);
 
     ap_hook_child_init(modperl_hook_child_init,
                        NULL, NULL, APR_HOOK_FIRST);
@@ -728,22 +876,24 @@ void modperl_register_hooks(apr_pool_t *p)
 
 static const command_rec modperl_cmds[] = {  
     MP_CMD_SRV_ITERATE("PerlSwitches", switches, "Perl Switches"),
-    MP_CMD_SRV_ITERATE("PerlModule", modules, "PerlModule"),
-    MP_CMD_SRV_ITERATE("PerlRequire", requires, "PerlRequire"),
+    MP_CMD_DIR_ITERATE("PerlModule", modules, "PerlModule"),
+    MP_CMD_DIR_ITERATE("PerlRequire", requires, "PerlRequire"),
+    MP_CMD_SRV_ITERATE("PerlConfigRequire", config_requires, "PerlConfigRequire"),
+    MP_CMD_SRV_ITERATE("PerlPostConfigRequire", post_config_requires, "PerlPostConfigRequire"),
     MP_CMD_DIR_ITERATE("PerlOptions", options, "Perl Options"),
     MP_CMD_DIR_ITERATE("PerlInitHandler", init_handlers, "Subroutine name"),
     MP_CMD_DIR_TAKE2("PerlSetVar", set_var, "PerlSetVar"),
     MP_CMD_DIR_ITERATE2("PerlAddVar", add_var, "PerlAddVar"),
     MP_CMD_DIR_TAKE2("PerlSetEnv", set_env, "PerlSetEnv"),
     MP_CMD_SRV_TAKE1("PerlPassEnv", pass_env, "PerlPassEnv"),
-    MP_CMD_SRV_RAW_ARGS_ON_READ("<Perl", perl, "Perl Code"),
-    MP_CMD_SRV_RAW_ARGS("Perl", perldo, "Perl Code"),
+    MP_CMD_DIR_RAW_ARGS_ON_READ("<Perl", perl, "Perl Code"),
+    MP_CMD_DIR_RAW_ARGS("Perl", perldo, "Perl Code"),
 
     MP_CMD_DIR_TAKE1("PerlSetInputFilter", set_input_filter,
                      "filter[;filter]"),
     MP_CMD_DIR_TAKE1("PerlSetOutputFilter", set_output_filter,
                      "filter[;filter]"),
-    
+
     MP_CMD_DIR_RAW_ARGS_ON_READ("=pod", pod, "Start of POD"),
     MP_CMD_DIR_RAW_ARGS_ON_READ("=back", pod, "End of =over"),
     MP_CMD_DIR_RAW_ARGS_ON_READ("=cut", pod_cut, "End of POD"),
@@ -838,11 +988,38 @@ static int modperl_response_handler_run(request_rec *r, int finish)
 
 int modperl_response_handler(request_rec *r)
 {
+    MP_dDCFG;
+    apr_status_t retval;
+
+#ifdef USE_ITHREADS
+    pTHX;
+    modperl_interp_t *interp;
+#endif
+
     if (!strEQ(r->handler, "modperl")) {
         return DECLINED;
     }
 
-    return modperl_response_handler_run(r, TRUE);
+#ifdef USE_ITHREADS
+    interp = modperl_interp_select(r, r->connection, r->server);
+    aTHX = interp->perl;
+#endif
+
+    /* default is -SetupEnv, add if PerlOption +SetupEnv */
+    if (MpDirSETUP_ENV(dcfg)) {
+        modperl_env_request_populate(aTHX_ r);
+    }
+
+    retval = modperl_response_handler_run(r, TRUE);
+
+#ifdef USE_ITHREADS
+    if (MpInterpPUTBACK(interp)) {
+        /* PerlInterpScope handler */
+        modperl_interp_unselect(interp);
+    }
+#endif
+
+    return retval;
 }
 
 int modperl_response_handler_cgi(request_rec *r)
@@ -922,6 +1099,17 @@ int modperl_response_handler_cgi(request_rec *r)
     return retval;
 }
 
+/* This ugly hack pulls in any function listed in
+ * modperl_exports.c. Otherwise, the over-zealous
+ * linker would remove unused api functions
+ */
+const void *modperl_suck_in_ugly_hack(void);
+const void *modperl_suck_in_ugly_hack(void)
+{
+    extern const void *modperl_ugly_hack;
+    return modperl_ugly_hack;
+}
+
 module AP_MODULE_DECLARE_DATA perl_module = {
     STANDARD20_MODULE_STUFF, 
     modperl_config_dir_create, /* dir config creater */
