@@ -1,4 +1,4 @@
-/* Copyright 2000-2004 The Apache Software Foundation
+/* Copyright 2000-2005 The Apache Software Foundation
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,6 +21,7 @@ int modperl_require_module(pTHX_ const char *pv, int logfailure)
 
     dSP;
     PUSHSTACKi(PERLSI_REQUIRE);
+    ENTER;SAVETMPS;
     PUTBACK;
     sv = sv_newmortal();
     sv_setpv(sv, "require ");
@@ -28,6 +29,7 @@ int modperl_require_module(pTHX_ const char *pv, int logfailure)
     eval_sv(sv, G_DISCARD);
     SPAGAIN;
     POPSTACK;
+    FREETMPS;LEAVE;
 
     if (SvTRUE(ERRSV)) {
         if (logfailure) {
@@ -36,7 +38,7 @@ int modperl_require_module(pTHX_ const char *pv, int logfailure)
         }
         return FALSE;
     }
-        
+
     return TRUE;
 }
 
@@ -79,17 +81,37 @@ static SV *modperl_hv_request_find(pTHX_ SV *in, char *classname, CV *cv)
         Perl_croak(aTHX_
                    "method `%s' invoked by a `%s' object with no `r' key!",
                    cv ? GvNAME(CvGV(cv)) : "unknown",
-                   HvNAME(SvSTASH(SvRV(in))));
+                   (SvRV(in) && SvSTASH(SvRV(in)))
+                       ? HvNAME(SvSTASH(SvRV(in)))
+                       : "unknown");
     }
 
     return SvROK(sv) ? SvRV(sv) : sv;
 }
 
+
+/* notice that if sv is not an Apache::ServerRec object and
+ * Apache->request is not available, the returned global object might
+ * be not thread-safe under threaded mpms, so use with care
+ */
+
 MP_INLINE server_rec *modperl_sv2server_rec(pTHX_ SV *sv)
 {
-    return SvOBJECT(sv) ?
-        (server_rec *)SvObjIV(sv) :
-        modperl_global_get_server_rec();
+    if (SvOBJECT(sv) || (SvROK(sv) && (SvTYPE(SvRV(sv)) == SVt_PVMG))) {
+        return (server_rec *)SvObjIV(sv);
+    }
+
+    /* next see if we have Apache->request available */
+    {
+        request_rec *r = NULL;
+        (void)modperl_tls_get_request_rec(&r);
+        if (r) {
+            return r->server;
+        }
+    }
+
+    /* modperl_global_get_server_rec is not thread safe w/o locking */
+    return modperl_global_get_server_rec();
 }
 
 MP_INLINE request_rec *modperl_sv2request_rec(pTHX_ SV *sv)
@@ -118,15 +140,12 @@ request_rec *modperl_xs_sv2request_rec(pTHX_ SV *in, char *classname, CV *cv)
         }
     }
 
-    if (!sv) {
+    /* might be Apache::ServerRec::warn method */
+    if (!sv && !(classname && SvPOK(in) && !strEQ(classname, SvPVX(in)))) {
         request_rec *r = NULL;
         (void)modperl_tls_get_request_rec(&r);
 
         if (!r) {
-            if (classname && SvPOK(in) && !strEQ(classname, SvPVX(in))) {
-                /* might be Apache::{Server,RequestRec}-> dual method */
-                return NULL;
-            }
             Perl_croak(aTHX_
                        "Apache->%s called without setting Apache->request!",
                        cv ? GvNAME(CvGV(cv)) : "unknown");
@@ -135,7 +154,9 @@ request_rec *modperl_xs_sv2request_rec(pTHX_ SV *in, char *classname, CV *cv)
         return r;
     }
 
-    if ((mg = mg_find(sv, PERL_MAGIC_ext))) {
+    /* there could be pool magic attached to custom $r object, so make
+     * sure that mg->mg_ptr is set */
+    if ((mg = mg_find(sv, PERL_MAGIC_ext)) && mg->mg_ptr) {
         return (request_rec *)mg->mg_ptr;
     }
     else {
@@ -179,88 +200,6 @@ MP_INLINE SV *modperl_ptr2obj(pTHX_ char *classname, void *ptr)
     return sv;
 }
 
-/* XXX: sv_setref_uv does not exist in 5.6.x */
-MP_INLINE SV *modperl_perl_sv_setref_uv(pTHX_ SV *rv,
-                                        const char *classname, UV uv)
-{
-    sv_setuv(newSVrv(rv, classname), uv);
-    return rv;
-}
-
-static apr_pool_t *modperl_sv2pool(pTHX_ SV *obj, CV *method)
-{
-    apr_pool_t *p = NULL;
-    char *classname = NULL;
-    IV ptr = 0;
-
-    if ((SvROK(obj) && (SvTYPE(SvRV(obj)) == SVt_PVMG))) {
-        /* standard classes */
-        classname = SvCLASS(obj);
-        ptr = SvObjIV(obj);
-    }
-    else if ((SvROK(obj) && (SvTYPE(SvRV(obj)) == SVt_PVHV))) {
-        /* Apache::RequestRec subclass */
-        classname = SvCLASS(obj);
-        ptr = SvIV(modperl_hv_request_find(aTHX_ obj, classname, method));
-
-        /* if modperl_hv_request_find succeeeds then the class is an 
-         * Apache::RequestRec subclass (the only subclass we support).
-         * so, fake things a bit so we can dig out the proper pool below
-         */
-         classname = "Apache::RequestRec";
-    }
-    else {
-        MP_TRACE_m(MP_FUNC, "SV not a recognized object");
-        return NULL;
-    }
-
-    if (strnEQ(classname, "APR::", 5)) {
-        classname += 5;
-        switch (*classname) {
-          case 'P':
-            if (strEQ(classname, "Pool")) {
-                p = (apr_pool_t *)SvObjIV(obj);
-            }
-            break;
-          default:
-            MP_TRACE_m(MP_FUNC, "class %s not recognized", classname);
-            break;
-        };
-    }
-    else if (strnEQ(classname, "Apache::", 8)) {
-        classname += 8;
-        switch (*classname) {
-          case 'C':
-            if (strEQ(classname, "Connection")) {
-                p = ((conn_rec *)ptr)->pool;
-            }
-            break;
-          case 'R':
-            if (strEQ(classname, "RequestRec")) {
-                p = ((request_rec *)ptr)->pool;
-            }
-            break;
-          case 'S':
-            if (strEQ(classname, "Server")) {
-                p = ((server_rec *)ptr)->process->pconf;
-            }
-            break;
-          default:
-            MP_TRACE_m(MP_FUNC, "class %s not recognised", classname);
-            break;
-        };
-    }
-    else {
-        MP_TRACE_m(MP_FUNC, "class %s not recognised", classname);
-    }
-
-    if (p == NULL) {
-        MP_TRACE_m(MP_FUNC, "unable to derive pool from object");
-    }
-
-    return p;
-}
-
 int modperl_errsv(pTHX_ int status, request_rec *r, server_rec *s)
 {
     SV *sv = ERRSV;
@@ -290,19 +229,22 @@ int modperl_errsv(pTHX_ int status, request_rec *r, server_rec *s)
     return status;
 }
 
-char *modperl_server_desc(server_rec *s, apr_pool_t *p)
+/* prepends the passed sprintf-like arguments to ERRSV, which also
+ * gets stringified on the way */
+void modperl_errsv_prepend(pTHX_ const char *pat, ...)
 {
-    return apr_psprintf(p, "%s:%u", s->server_hostname, s->port);
+    SV *sv;
+    va_list args;
+
+    va_start(args, pat);
+    sv = vnewSVpvf(pat, &args);
+    va_end(args);
+
+    sv_catsv(sv, ERRSV);
+    sv_copypv(ERRSV, sv);
+    sv_free(sv);
 }
 
-/* used in debug traces */
-MP_INLINE char *modperl_pid_tid(apr_pool_t *p)
-{
-    return apr_psprintf(p, "%lu" MP_TRACEf_TID,
-                        (unsigned long)getpid() MP_TRACEv__TID);
-}
-
-    
 #define dl_librefs "DynaLoader::dl_librefs"
 #define dl_modules "DynaLoader::dl_modules"
 
@@ -379,6 +321,59 @@ void modperl_xs_dl_handles_close(void **handles)
     free(handles);
 }
 
+/* XXX: There is no XS accessible splice() */
+static void modperl_av_remove_entry(pTHX_ AV *av, I32 index)
+{
+    I32 i;
+    AV *tmpav = newAV();
+
+    /* stash the entries _before_ the item to delete */
+    for (i=0; i<=index; i++) {
+        av_store(tmpav, i, SvREFCNT_inc(av_shift(av)));
+    }
+
+    /* make size at the beginning of the array */
+    av_unshift(av, index-1);
+
+    /* add stashed entries back */
+    for (i=0; i<index; i++) {
+        av_store(av, i, *av_fetch(tmpav, i, 0));
+    }
+
+    sv_free((SV *)tmpav);
+}
+
+static void modperl_package_unload_dynamic(pTHX_ const char *package, 
+                                           I32 dl_index)
+{
+    AV *librefs = get_av(dl_librefs, 0);
+    SV *libref = *av_fetch(librefs, dl_index, 0);
+
+    modperl_sys_dlclose((void *)SvIV(libref));
+
+    /* remove package from @dl_librefs and @dl_modules */
+    modperl_av_remove_entry(aTHX_ get_av(dl_librefs, 0), dl_index);
+    modperl_av_remove_entry(aTHX_ get_av(dl_modules, 0), dl_index);
+
+    return;    
+}
+
+static int modperl_package_is_dynamic(pTHX_ const char *package,
+                                      I32 *dl_index)
+{
+   I32 i;
+   AV *modules = get_av(dl_modules, FALSE);
+
+   for (i=0; i<av_len(modules); i++) {
+        SV *module = *av_fetch(modules, i, 0);
+        if (strEQ(package, SvPVX(module))) {
+            *dl_index = i;
+            return TRUE;
+        }
+    }
+    return FALSE;
+}
+
 modperl_cleanup_data_t *modperl_cleanup_data_new(apr_pool_t *p, void *data)
 {
     modperl_cleanup_data_t *cdata =
@@ -388,61 +383,6 @@ modperl_cleanup_data_t *modperl_cleanup_data_new(apr_pool_t *p, void *data)
     return cdata;
 }
 
-MP_INLINE modperl_uri_t *modperl_uri_new(apr_pool_t *p)
-{
-    modperl_uri_t *uri = (modperl_uri_t *)apr_pcalloc(p, sizeof(*uri));
-    uri->pool = p;
-    return uri;
-}
-
-MP_INLINE SV *modperl_hash_tie(pTHX_ 
-                               const char *classname,
-                               SV *tsv, void *p)
-{
-    SV *hv = (SV*)newHV();
-    SV *rsv = sv_newmortal();
-
-    sv_setref_pv(rsv, classname, p);
-    sv_magic(hv, rsv, PERL_MAGIC_tied, Nullch, 0);
-
-    return SvREFCNT_inc(sv_bless(sv_2mortal(newRV_noinc(hv)),
-                                 gv_stashpv(classname, TRUE)));
-}
-
-MP_INLINE void *modperl_hash_tied_object(pTHX_ 
-                                         const char *classname,
-                                         SV *tsv)
-{
-    if (sv_derived_from(tsv, classname)) {
-        if (SVt_PVHV == SvTYPE(SvRV(tsv))) {
-            SV *hv = SvRV(tsv);
-            MAGIC *mg;
-
-            if (SvMAGICAL(hv)) {
-                if ((mg = mg_find(hv, PERL_MAGIC_tied))) {
-                    return (void *)MgObjIV(mg);
-                }
-                else {
-                    Perl_warn(aTHX_ "Not a tied hash: (magic=%c)", mg);
-                }
-            }
-            else {
-                Perl_warn(aTHX_ "SV is not tied");
-            }
-        }
-        else {
-            return (void *)SvObjIV(tsv);
-        }
-    }
-    else {
-        Perl_croak(aTHX_
-                   "argument is not a blessed reference "
-                   "(expecting an %s derived object)", classname);
-    }
-
-    return NULL;
-}
-
 MP_INLINE void modperl_perl_av_push_elts_ref(pTHX_ AV *dst, AV *src)
 {
     I32 i, j, src_fill = AvFILLp(src), dst_fill = AvFILLp(dst);
@@ -533,7 +473,7 @@ void modperl_perl_call_list(pTHX_ AV *subs, const char *name)
                " running %d %s subs",
                (unsigned long)getpid(), MP_TRACEv_TID_ MP_TRACEv_PERLID_
                AvFILLp(subs)+1, name);
-    
+
     for (i=0; i<=AvFILLp(subs); i++) {
 	CV *cv = (CV*)ary[i];
 	SV *atsv = ERRSV;
@@ -582,7 +522,7 @@ MP_INLINE SV *modperl_dir_config(pTHX_ request_rec *r, server_rec *s,
             retval = &PL_sv_undef;
         }
     }
-        
+
     return retval;
 }
 
@@ -620,73 +560,41 @@ SV *modperl_table_get_set(pTHX_ apr_table_t *table, char *key,
     return retval;
 }
 
+static char *package2filename(const char *package, int *len)
+{
+    const char *s;
+    char *d;
+    char *filename;
+
+    filename = malloc((strlen(package)+4)*sizeof(char));
+
+    for (s = package, d = filename; *s; s++, d++) {
+        if (*s == ':' && s[1] == ':') {
+            *d = '/';
+            s++;
+        }
+        else {
+            *d = *s;
+        }
+    }
+    *d++ = '.';
+    *d++ = 'p';
+    *d++ = 'm';
+    *d   = '\0';
+
+    *len = d - filename;
+    return filename;
+}
+
 MP_INLINE int modperl_perl_module_loaded(pTHX_ const char *name)
 {
-    return (*name && gv_stashpv(name, FALSE)) ? 1 : 0;
-}
+    SV **svp;
+    int len;
+    char *filename = package2filename(name, &len);
+    svp = hv_fetch(GvHVn(PL_incgv), filename, len, 0);
+    free(filename);
 
-/* same as Symbol::gensym() */
-SV *modperl_perl_gensym(pTHX_ char *pack)
-{
-    GV *gv = newGVgen(pack);
-    SV *rv = newRV((SV*)gv);
-    (void)hv_delete(gv_stashpv(pack, TRUE), 
-                    GvNAME(gv), GvNAMELEN(gv), G_DISCARD);
-    return rv;
-}
-
-static int modperl_gvhv_is_stash(GV *gv)
-{
-    int len = GvNAMELEN(gv);
-    char *name = GvNAME(gv);
-
-    if ((len > 2) && (name[len - 1] == ':') && (name[len - 2] == ':')) {
-        return 1;
-    }
-
-    return 0;
-}
-
-/*
- * we do not clear symbols within packages, the desired behavior
- * for directive handler classes.  and there should never be a package
- * within the %Apache::ReadConfig.  nothing else that i'm aware of calls
- * this function, so we should be ok.
- */
-
-void modperl_clear_symtab(pTHX_ HV *symtab) 
-{
-    SV *val;
-    char *key;
-    I32 klen;
-
-    hv_iterinit(symtab);
-    
-    while ((val = hv_iternextsv(symtab, &key, &klen))) {
-        SV *sv;
-        HV *hv;
-        AV *av;
-        CV *cv;
-
-        if ((SvTYPE(val) != SVt_PVGV) || GvIMPORTED((GV*)val)) {
-            continue;
-        }
-        if ((sv = GvSV((GV*)val))) {
-            sv_setsv(GvSV((GV*)val), &PL_sv_undef);
-        }
-        if ((hv = GvHV((GV*)val)) && !modperl_gvhv_is_stash((GV*)val)) {
-            hv_clear(hv);
-        }
-        if ((av = GvAV((GV*)val))) {
-            av_clear(av);
-        }
-        if ((cv = GvCV((GV*)val)) && (GvSTASH((GV*)val) == GvSTASH(CvGV(cv)))) {
-            GV *gv = CvGV(cv);
-            cv_undef(cv);
-            CvGV(cv) = gv;
-            GvCVGEN(gv) = 1; /* invalidate method cache */
-        }
-    }
+    return (svp && *svp != &PL_sv_undef) ? 1 : 0;
 }
 
 #define SLURP_SUCCESS(action) \
@@ -702,7 +610,7 @@ MP_INLINE SV *modperl_slurp_filename(pTHX_ request_rec *r, int tainted)
     apr_status_t rc;
     apr_size_t size;
     apr_file_t *file;
-    
+
     size = r->finfo.size;
     sv = newSV(size);
 
@@ -724,7 +632,7 @@ MP_INLINE SV *modperl_slurp_filename(pTHX_ request_rec *r, int tainted)
     SLURP_SUCCESS("reading");
 
     MP_TRACE_o(MP_FUNC, "read %d bytes from '%s'\n", size, r->filename);
-    
+
     if (r->finfo.size != size) {
         SvREFCNT_dec(sv); 
         Perl_croak(aTHX_ "Error: read %d bytes, expected %d ('%s')",
@@ -733,7 +641,7 @@ MP_INLINE SV *modperl_slurp_filename(pTHX_ request_rec *r, int tainted)
 
     rc = apr_file_close(file);
     SLURP_SUCCESS("closing");
-    
+
     SvPVX(sv)[size] = '\0';
     SvCUR_set(sv, size);
     SvPOK_on(sv);
@@ -744,7 +652,7 @@ MP_INLINE SV *modperl_slurp_filename(pTHX_ request_rec *r, int tainted)
     else {
         SvTAINTED_off(sv);
     }
-    
+
     return newRV_noinc(sv);
 }
 
@@ -783,7 +691,7 @@ char *modperl_file2package(apr_pool_t *p, const char *file)
             while (*(file+1) && MP_VALID_PATH_DELIM(*(file+1))) {
                 file++;
             }
- 
+
             /* path delim not until end of line */
             if (*(file+1)) {
                 *c = *(c+1) = ':';
@@ -794,138 +702,133 @@ char *modperl_file2package(apr_pool_t *p, const char *file)
             *c = '_';
         }
     }
-   
+
     return package;
 }
 
-/* this is used across server_root_relative() in the
- * Apache, Apache::Server, Apache::RequestRec, and 
- * Apache::Connection classes
- */
-SV *modperl_server_root_relative(pTHX_ SV *sv, const char *fname)
+SV *modperl_apr_array_header2avrv(pTHX_ apr_array_header_t *array)
 {
-    apr_pool_t *p;
+    AV *av = newAV(); 
 
-    if (!sv_isobject(sv)) {
-        Perl_croak(aTHX_ "usage: Apache::server_root_relative(obj, name)");
-    }
-
-    p = modperl_sv2pool(aTHX_ sv, get_cv("Apache::server_root_relative", 0));
-
-    if (p == NULL) {
-        MP_TRACE_a(MP_FUNC,
-                   "unable to isolate pool for ap_server_root_relative()");
-        return &PL_sv_undef;
-    }
-
-    /* copy the SV in case the pool goes out of scope before the perl scalar */
-    return newSVpv(ap_server_root_relative(p, fname), 0);
-}
-
-char *modperl_coderef2text(pTHX_ apr_pool_t *p, CV *cv)
-{
-    dSP;
-    int count;
-    SV *bdeparse;
-    char *text;
-    
-    /* B::Deparse >= 0.61 needed for blessed code references.
-     * 0.6 works fine for non-blessed code refs.
-     * notice that B::Deparse is not CPAN-updatable.
-     * 0.61 is available starting from 5.8.0
-     */
-    load_module(PERL_LOADMOD_NOIMPORT,
-                newSVpvn("B::Deparse", 10),
-                newSVnv(SvOBJECT((SV*)cv) ? 0.61 : 0.60));
-
-    ENTER;
-    SAVETMPS;
-
-    /* create the B::Deparse object */
-    PUSHMARK(sp);
-    XPUSHs(sv_2mortal(newSVpvn("B::Deparse", 10)));
-    PUTBACK;
-    count = call_method("new", G_SCALAR);
-    SPAGAIN;
-    if (count != 1) {
-        Perl_croak(aTHX_ "Unexpected return value from B::Deparse::new\n");
-    }
-    if (SvTRUE(ERRSV)) {
-        Perl_croak(aTHX_ "error: %s", SvPVX(ERRSV));
-    }
-    bdeparse = POPs;
-
-    PUSHMARK(sp);
-    XPUSHs(bdeparse);
-    XPUSHs(sv_2mortal(newRV_inc((SV*)cv)));
-    PUTBACK;
-    count = call_method("coderef2text", G_SCALAR);
-    SPAGAIN;
-    if (count != 1) {
-        Perl_croak(aTHX_ "Unexpected return value from "
-                   "B::Deparse::coderef2text\n");
-    }
-    if (SvTRUE(ERRSV)) {
-        Perl_croak(aTHX_ "error: %s", SvPVX(ERRSV));
-    }
-    
-    {
-        STRLEN n_a;
-        text = apr_pstrcat(p, "sub ", POPpx, NULL);
-    }
-    
-    PUTBACK;
-    
-    FREETMPS;
-    LEAVE;
-
-    return text;
-}
-
-#ifdef MP_TRACE
-
-/* XXX: internal debug function, a candidate for modperl_debug.c */
-/* any non-false value for MOD_PERL_TRACE/PerlTrace enables this function */
-void modperl_apr_table_dump(pTHX_ apr_table_t *table, char *name)
-{
-    int i;
-    const apr_array_header_t *array;
-    apr_table_entry_t *elts;
-
-    array = apr_table_elts(table);
-    elts  = (apr_table_entry_t *)array->elts;
-    modperl_trace(MP_FUNC, "Contents of table %s", name);
-    for (i = 0; i < array->nelts; i++) {
-        if (!elts[i].key || !elts[i].val) {
-            continue;
+    if (array) {
+        int i; 
+        for (i = 0; i < array->nelts; i++) {
+            av_push(av, newSVpv(((char **)array->elts)[i], 0));
         }
-        modperl_trace(MP_FUNC, "%s => %s", elts[i].key, elts[i].val);
+    }
+    return newRV_noinc((SV*)av);
+}
+
+apr_array_header_t *modperl_avrv2apr_array_header(pTHX_ apr_pool_t *p,
+                                                  SV *avrv)
+{
+    AV *av;
+    apr_array_header_t *array;
+    int i, av_size;
+
+    if (!(SvROK(avrv) && (SvTYPE(SvRV(avrv)) == SVt_PVAV))) {
+        Perl_croak(aTHX_ "Not an array reference");
+    }
+
+    av = (AV*)SvRV(avrv);
+    av_size = av_len(av);
+    array = apr_array_make(p, av_size+1, sizeof(char *));
+
+    for (i = 0; i <= av_size; i++) {
+        SV *sv = *av_fetch(av, i, FALSE);
+        char **entry = (char **)apr_array_push(array);
+        *entry = apr_pstrdup(p, SvPV(sv, PL_na));
+    }
+
+    return array;
+}
+
+/* Remove a package from %INC */
+static void modperl_package_delete_from_inc(pTHX_ const char *package)  
+{
+    int len;
+    char *filename = package2filename(package, &len);
+    hv_delete(GvHVn(PL_incgv), filename, len, G_DISCARD);
+    free(filename);
+}
+
+/* Destroy a package's stash */
+#define MP_STASH_SUBSTASH(key, len) ((len >= 2) &&                  \
+                                     (key[len-1] == ':') &&         \
+                                     (key[len-2] == ':'))   
+#define MP_STASH_DEBUGGER(key, len) ((len >= 2) &&                  \
+                                     (key[0] == '_') &&             \
+                                     (key[1] == '<'))
+#define MP_SAFE_STASH(key, len)     (!(MP_STASH_SUBSTASH(key,len)|| \
+                                      (MP_STASH_DEBUGGER(key, len))))
+static void modperl_package_clear_stash(pTHX_ const char *package)
+{
+    HV *stash;
+    if ((stash = gv_stashpv(package, FALSE))) {
+        HE *he;
+        I32 len;
+        char *key;
+        hv_iterinit(stash);
+        while ((he = hv_iternext(stash))) {
+            key = hv_iterkey(he, &len);
+            if (MP_SAFE_STASH(key, len)) {
+                SV *val = hv_iterval(stash, he);
+                char *this_stash = HvNAME(GvSTASH(val));
+                /* The safe thing to do is to skip over stash entries
+                 * that don't come from the package we are trying to
+                 * unload
+                 */
+                if (strcmp(this_stash, package) == 0) {
+                    hv_delete(stash, key, len, G_DISCARD);
+                }
+            }
+        }
+    }
+}
+
+/* Unload a module as completely and cleanly as possible */
+void modperl_package_unload(pTHX_ const char *package)
+{
+    I32 dl_index;
+
+    modperl_package_clear_stash(aTHX_ package);
+    modperl_package_delete_from_inc(aTHX_ package);
+
+    if (modperl_package_is_dynamic(aTHX_ package, &dl_index)) {
+        modperl_package_unload_dynamic(aTHX_ package, dl_index);
+    }
+
+}
+
+#define MP_RESTART_COUNT_KEY "mod_perl_restart_count"
+
+/* passing the main server object here, just because we don't have the
+ * modperl_server_pool available yet, later on we can access it
+ * through the modperl_server_pool() call.
+ */
+void modperl_restart_count_inc(server_rec *base_server)
+{
+    void *data;
+    int *counter;
+    apr_pool_t *p = base_server->process->pool;
+
+    apr_pool_userdata_get(&data, MP_RESTART_COUNT_KEY, p);
+    if (data) {
+        counter = data;
+        (*counter)++;
+    }
+    else {
+        counter = apr_palloc(p, sizeof *counter);
+        *counter = 1;
+        apr_pool_userdata_set(counter, MP_RESTART_COUNT_KEY,
+                              apr_pool_cleanup_null, p);
     }    
 }
 
-/* XXX: internal debug function, a candidate for modperl_debug.c */
-void modperl_perl_modglobal_dump(pTHX)
+int modperl_restart_count(void)
 {
-    HV *hv = PL_modglobal;
-    AV *val;
-    char *key;
-    I32 klen;
-    hv_iterinit(hv);
-
-    MP_TRACE_g(MP_FUNC, "|-------- PL_modglobal --------");
-#ifdef USE_ITHREADS
-    MP_TRACE_g(MP_FUNC, "| perl 0x%lx", (unsigned long)aTHX);
-#endif
-    MP_TRACE_g(MP_FUNC, "| PL_modglobal 0x%lx",
-               (unsigned long)PL_modglobal);
-    
-    while ((val = (AV*)hv_iternextsv(hv, &key, &klen))) {
-        MP_TRACE_g(MP_FUNC, "| %s => 0x%lx", key, val);
-    }
-    
-    MP_TRACE_g(MP_FUNC, "|-------- PL_modglobal --------\n");
-        
-}
-
-
-#endif
+    void *data;
+    apr_pool_userdata_get(&data, MP_RESTART_COUNT_KEY,
+                          modperl_global_get_server_rec()->process->pool);
+    return data ? *(int *)data : 0;
+ }
