@@ -10,47 +10,27 @@ use Apache::RequestIO ();
 use APR::Brigade ();
 use APR::Bucket ();
 
+use TestCommon::Utils ();
+
 use Apache::Const -compile => qw(OK M_POST);
 use APR::Const -compile => ':common';
 
 sub handler : FilterRequestHandler {
     my($filter, $bb, $mode, $block, $readbytes) = @_;
 
-    #warn "Called!";
-    my $ba = $filter->r->connection->bucket_alloc;
+    $filter->next->get_brigade($bb, $mode, $block, $readbytes);
 
-    my $ctx_bb = APR::Brigade->new($filter->r->pool, $ba);
+    for (my $b = $bb->first; $b; $b = $bb->next($b)) {
 
-    my $rv = $filter->next->get_brigade($ctx_bb, $mode, $block, $readbytes);
+        last if $b->is_eos;
 
-    if ($rv != APR::SUCCESS) {
-        return $rv;
-    }
-
-    while (!$ctx_bb->empty) {
-        my $data;
-        my $bucket = $ctx_bb->first;
-
-        $bucket->remove;
-
-        if ($bucket->is_eos) {
-            #warn "EOS!!!!";
-            $bb->insert_tail($bucket);
-            last;
-        }
-
-        my $status = $bucket->read($data);
-        #warn "DATA bucket!!!!";
-        if ($status != APR::SUCCESS) {
-            return $status;
-        }
-
-        if ($data) {
+        if ($b->read(my $data)) {
             #warn"[$data]\n";
-            $bucket = APR::Bucket->new(scalar reverse $data);
+            my $nb = APR::Bucket->new($bb->bucket_alloc, scalar reverse $data);
+            $b->insert_before($nb);
+            $b->delete;
+            $b = $nb;
         }
-
-        $bb->insert_tail($bucket);
     }
 
     Apache::OK;
@@ -62,7 +42,7 @@ sub response {
     $r->content_type('text/plain');
 
     if ($r->method_number == Apache::M_POST) {
-        my $data = ModPerl::Test::read_post($r);
+        my $data = TestCommon::Utils::read_post($r);
         $r->puts($data);
     }
     else {
