@@ -22,45 +22,27 @@ sub handler : FilterConnectionHandler {
     my($filter, $bb, $mode, $block, $readbytes) = @_;
 
     debug "FILTER CALLED";
-    my $c = $filter->c;
-    my $ctx_bb = APR::Brigade->new($c->pool, $c->bucket_alloc);
 
-    my $rv = $filter->next->get_brigade($ctx_bb, $mode, $block, $readbytes);
+    $filter->next->get_brigade($bb, $mode, $block, $readbytes);
 
-    if ($rv != APR::SUCCESS) {
-        return $rv;
-    }
+    for (my $b = $bb->first; $b; $b = $bb->next($b)) {
 
-    while (!$ctx_bb->empty) {
-        my $data;
-        my $bucket = $ctx_bb->first;
+        last if $b->is_eos;
 
-        $bucket->remove;
-
-        if ($bucket->is_eos) {
-            debug "EOS!!!";
-            $bb->insert_tail($bucket);
-            last;
-        }
-
-        my $status = $bucket->read($data);
-        debug "FILTER READ:\n$data";
-
-        if ($status != APR::SUCCESS) {
-            return $status;
-        }
-
-        if ($data and $data =~ s,GET $from_url,GET $to_url,) {
+        if ($b->read(my $data)) {
+            next unless $data =~ s|GET $from_url|GET $to_url|;
             debug "GET line rewritten to be:\n$data";
-            $bucket = APR::Bucket->new($data);
-            # XXX: currently a bug in httpd doesn't allow to remove
-            # the first connection filter. once it's fixed adjust the test
-            # to test that it was invoked only once.
-            # debug "removing the filter";
-            # $filter->remove; # this filter is no longer needed
+            my $nb = APR::Bucket->new($bb->bucket_alloc, $data);
+            $b->insert_before($nb);
+            $b->delete;
+            $b = $nb;
         }
 
-        $bb->insert_tail($bucket);
+        # XXX: currently a bug in httpd doesn't allow to remove
+        # the first connection filter. once it's fixed adjust the test
+        # to test that it was invoked only once.
+        # debug "removing the filter";
+        # $filter->remove; # this filter is no longer needed
     }
 
     Apache::OK;
