@@ -23,7 +23,6 @@ sub handler {
 
     debug join '', "-" x 20 , " filter called ", "-" x 20;
 
-    t_server_log_error_is_expected();
     die "This filter must die";
 
     return Apache::OK;
@@ -32,10 +31,14 @@ sub handler {
 sub response {
     my $r = shift;
 
+    # cause taint problems, as there was a bug (panic: POPSTACK)
+    # caused when APR/Error.pm was attempted to be loaded from
+    # $r->read() when the latter was trying to croak about the failed
+    # read, due to the filter returning 500
+    eval { system('echo', 'hello') };
+
+    t_server_log_error_is_expected(2);
     my $len = $r->read(my $data, $r->headers_in->{'Content-Length'});
-    t_server_log_error_is_expected();
-    die "failed to read POSTed data: $!" unless defined $len;
-    debug "read $len bytes [$data]";
 
     $r->content_type('text/plain');
     $r->print("it shouldn't be printed, because the input filter has died");
