@@ -15,11 +15,6 @@ sub handler {
         die "\$ENV{MOD_PERL} is not set";
     }
 
-    my $gw = $ENV{GATEWAY_INTERFACE} || '';
-    unless ($gw eq 'CGI-Perl/1.1') {
-        die "\$ENV{GATEWAY_INTERFACE} is not properly set ($gw)";
-    }
-
     if ($CGI::Q) {
         die "CGI.pm globals were not reset";
     }
@@ -55,4 +50,4 @@ sub handler {
 1;
 __END__
 SetHandler perl-script
-
+PerlOptions -SetupEnv
